from dataclasses import dataclass, field
import datetime as dt
import json
from pathlib import Path
from .config import get_config, update_config
import pandas as pd
import numpy as np
import importlib.util
import yaml
from .utils import (write_table,
                   read_table,
                   read_metadata,
                   write_yaml,
                   path_exists,
                   update_yaml,
                   date2iso,
                   set_path)
import importlib


cfg = get_config()
NOW = date2iso(str(dt.datetime.now()))
BASE_DIR = Path(__file__).resolve().parent.parent

@dataclass
class LocalDataset(object):
    """LocalDataset [summary]

    [extended_summary]

    Args:
        dataset_name ([str]): [description]
        datasets ([list]): [description]
        pipeline ([list]): [description]
        domain ([str]): [description]
        module ([list]): [description]
        dependencies ([list]): [description]
        since ([list]): [description]
        until ([list]): [description]
        delay ([list]): [description]
        reload ([list]): [description]
        reapply ([list]): [description]
        update ([list]): [description]
        home ([Path]): [description]
        compression ([str]): [description]
        data_folder ([str]): [description]
        data ([pd.DataFrame]): [description]

    Returns:
        [type]: [description]
    """
    dataset_name: str
    datasets: dict
    pipeline: list
    domain:  str
    module: str = 'base'
    dependencies: list = field(default_factory=lambda: [])
    since: str = '2019-04-01'
    until: str = NOW
    delay: float = 1
    reapply: bool = False
    reload: bool = False
    update: bool = False
    home: Path = cfg['home']
    compression: str = cfg['compression']
    data_folder: str = cfg['data_folder']
    data: pd.DataFrame = field(default_factory=lambda: pd.DataFrame())

    def __post_init__(self):
        """__post_init__ [summary]

        [extended_summary]
        """
        self.dataset_name = self.dataset_name.lower()
        self.domain = self.domain.lower()
        self._delay = dt.timedelta(hours=self.delay)
        self.since = date2iso(self.since)         
        try: 
            file_path = list(Path(BASE_DIR).rglob(self.module+'.py'))
            if len(file_path)==0:
                file_path = list(Path().rglob(self.module+'.py'))
        except:
            print('UNABLE to find module path')        
        spec = importlib.util.spec_from_file_location(self.module, file_path[0])
        self._module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(self._module)
        self.local_file = (f'{self.data_folder}/'
                           f'{self.domain}/'
                           f'{self.dataset_name}.parquet')
        self.metadata = {'since': self.since,
                         'until': self.until,
                         'Mac': cfg['mac']}
        self.load_dataset()
        self.data = read_table(self.local_file)

    def load_dataset(self):
        """load_dataset [summary]

        [extended_summary]
        """
        if not path_exists(self.local_file) or self.reload:
            set_path(self.local_file)
            self.process_dataset()
        elif self.update:
            self.update_dataset()
        elif self.reapply:
            self.reapply_dataset()    
        else:
            self.data = read_table(self.local_file)    

    def reapply_dataset(self):
        """reapply_dataset [summary]

        [extended_summary]
        """
        for func in self.pipeline:
            self.data = getattr(self._module, func)(self)
        self.save_dataset()
    
    def process_dataset(self):
        """process_dataset [summary]

        [extended_summary]
        """
        for func in self.pipeline:
            self.data = getattr(self._module, func)(self)
        domains = pd.DataFrame(cfg['domains'])
        dataset = np.array([self.domain,self.dataset_name])
        dataset_exist = (domains == dataset).all(axis=1).any()
        if not dataset_exist:
            self.register_dataset()
        self.save_dataset()    

    def update_dataset(self):
        """update_dataset [summary]

        [extended_summary]
        """
        hdr = read_metadata(self.local_file)
        metadata = json.loads(hdr.metadata[b'minder'].decode())
        until = pd.to_datetime(metadata['until']).tz_localize(None) + self._delay
        if until < pd.to_datetime(dt.datetime.now()):
            self.update_metadata()
            self.process_dataset()
            
            
    def update_metadata(self):
        """update_metadata [summary]

        [extended_summary]
        """
        if path_exists(self.local_file):
            since = self.load_metadata()['since']
        else:
            since = self.since
        self.metadata = {'since': since,
                         'until': self.until,
                         'Mac': cfg['mac']}            
    
    def load_metadata(self):
        """load_metadata [summary]

        [extended_summary]

        Returns:
            [type]: [description]
        """
        hdr = read_metadata(self.local_file)
        metadata = json.loads(hdr.metadata[b'minder'].decode())
        return metadata
    
            
    def save_dataset(self) -> None:
        """save_dataset [summary]

        [extended_summary]
        """
        write_table(self.data,
                    self.local_file,
                    self.compression,
                    self.metadata)

    def register_dataset(self) -> None:
        """register_dataset [summary]

        [extended_summary]
        """
        cfg['domains'].append({'domain':self.domain.lower(),'dataset':self.dataset_name.lower()})
        cfg['domains'] = pd.DataFrame(cfg['domains']).drop_duplicates().to_dict('records')
        update_config(cfg)
        home = cfg['home']
        dependencies = (pd.DataFrame(self.dependencies).
                          iloc[:,::-1].
                          rename(columns={1:'domain',0:'dataset'}).
                          to_dict(orient='records'))
        collection_file = f'{home}/dcarte/config/{self.domain}.yaml'
        data = {self.dataset_name: {"domains": dependencies,
                                    "pipeline": self.pipeline,
                                    "module": self.module}}
        if path_exists(collection_file):
            update_yaml(collection_file, data)
        else:
            write_yaml(collection_file, data)




from pathlib import Path
import shutil
import uuid
import os
import webbrowser
from .utils import (load_yaml,
                    write_yaml,
                    update_yaml,
                    path_exists)


def get_config(config_file : str = '/dcarte/config.yaml',
               root: Path = Path('__file__').parent.absolute(),
               home: Path = Path('~').expanduser()) -> dict:
    """get_config a function that returns or creates and returns a local config file 

    
    Args:
        config_file (str, optional): [description]. Defaults to '/dcarte/config.yaml'.
        root (Path, optional): [description]. Defaults to Path('__file__').parent.absolute().
        home (Path, optional): [description]. Defaults to Path('~').expanduser().

    Returns:
        [dict]: containing all the configuration information neeeded for dcarte 
    """
    if path_exists(str(home)+config_file):
        cfg = load_yaml(str(home)+config_file)
    else:
        cfg =  create_config(home, root)
    os.environ['MINDER_TOKEN'] = cfg['token']
    cfg.pop('token', None)
    return cfg


def update_config(new_dict:dict, home:Path = Path('~').expanduser()):
    """update_config updates the central config file with data from new_dict

    Args:
        new_dict (dict): [description]
        home (Path, optional): [description]. Defaults to Path('~').expanduser().
    """
    update_yaml(f"{home}/dcarte/config.yaml", new_dict)


def create_config(home:Path,root:Path):
    """create_config creates a baseline config file

    Args:
        home (Path): [description]
        root (Path): [description]

    Returns:
        [type]: [description]
    """
    # Create dcarte folder at home to store config and data folders
    for p in ["/dcarte/config", "/dcarte/data"]:
        Path(f"{home}{p}").mkdir(parents=True, exist_ok=True)
    # copy yaml files from source_yaml to home/config
    for p in Path(root).rglob('source_yaml'):
        if p.is_dir():
            source_yaml = p.resolve()
    try:        
        files = list(Path(source_yaml).glob('*.yaml'))
        [shutil.copy2(f'{file}', f'{home}/dcarte/config') for file in files]
    except:
        raise Exception("Sorry, nunable to copy base config yaml files")           
    # create a baseline config dict
    cfg = baseline_config(home,root,files)
    # open webpage and request user to copy token
    cfg['token'] = get_token()
    cfg['mac'] = get_mac()
    write_yaml(f"{home}/dcarte/config.yaml", cfg)
 
    return cfg
    

def get_mac() -> str:
    """get_mac return mac address of the compute node or computer

    Returns:
        str: [description]
    """
    return hex(uuid.getnode())


def get_token() -> str:
    """get_token opens the access-tokens website to create a unique REST token 

    Returns:
        str: a token generated at https://research.minder.care/portal/access-tokens
    """
    webbrowser.open('https://research.minder.care/portal/access-tokens')
    print('Please generate a token and copy it into the input bar')
    token = input()
    return token
    

def baseline_config(home: Path, root: Path, files:list) -> dict:
    """baseline_config create a baseline config dict 

    Args:
        home (Path): [description]
        root (Path): [description]

    Returns:
        dict: [description]
    """
        
    dataset_yamels = {file.stem:load_yaml(file) for file in files}
    datasets = [{'domain':domain, 'dataset': dataset} for domain, d in dataset_yamels.items() for dataset in d.keys()]
    headers = {'Accept': 'text/plain','Content-type': 'application/json'}
    cfg = {
        'compression': 'GZIP',
        'data_folder': f"{home}/dcarte/data",
        'domains': datasets,
        'headers': headers,
        'home': f'{home}',
        'root': f'{root}',
        'server': 'https://research.minder.care/api/export'
    }
    return cfg
    



import pandas as pd 
import numpy as np
from .local import LocalDataset
from .minder import MinderDataset
from .utils import (load_yaml, 
                    timer,
                    date2iso,
                    path_exists,
                    read_table)
from .config import get_config
import datetime as dt

NOW = date2iso(str(dt.datetime.now()))


@timer('Loading')
def load(dataset:str,domain:str,**kwargs):
    """load [summary]

    [extended_summary]

    Args:
        dataset (str): [description]
        domain (str): [description]

    Raises:
        Exception: [description]

    Returns:
        [type]: [description]
    """
    dataset,domain = dataset.lower(),domain.lower()
    cfg = get_config()
    dflt = get_defaults(**kwargs) | kwargs
    home = cfg['home']
    data_folder = cfg['data_folder']
    datasets = pd.DataFrame(cfg['domains'])
    if not (datasets == np.array([dataset,domain])).all(axis=1).any():
        raise Exception(f"Sorry, {dataset} is not a registered dataset in {domain} domain in dcarte")
    
    local_file = f'{data_folder}/{domain}/{dataset}.parquet'
    if path_exists(local_file) and not (dflt['update'] or dflt['reapply'] or dflt['reload']):
        return read_table(local_file)
    else:     
        info = load_yaml(f'{home}/dcarte/config/{domain}.yaml')
        if domain in ['raw','lookup']:
            input = {'dataset_name':dataset,
                     'datasets':info[dataset]['datasets'],
                     'columns':info[dataset]['columns'],
                     'dtypes':info[dataset]['dtype'],
                     'domain':domain} | dflt

            output = MinderDataset(**input)
        else:
            dependencies = pd.DataFrame(info[dataset]['domains'])
            parent_datasets = {}
            for _,row in dependencies.iterrows():
                parent_datasets[row.dataset] = load(row.dataset,row.domain, **dflt)
            input = {'dataset_name':dataset,
                     'datasets':parent_datasets,
                     'pipeline':info[dataset]['pipeline'],
                     'module':info[dataset]['module'],
                     'domain':domain} | dflt
            
            output = LocalDataset(**input)

        return output.data

def get_defaults(**kwargs):
    """get_defaults [summary]

    [extended_summary]

    Returns:
        [type]: [description]
    """
    cfg = get_config()
    defaults = {'since': '2019-04-01',
                'until': NOW,
                'delay': 1,
                'home': cfg['home'],
                'compression': cfg['compression'],
                'data_folder':  cfg['data_folder'],
                'update': False,
                'reload':  False,
                'reapply': False}

    defaults.update(kwargs)
    return defaults




from .config import get_config
from .utils import isnotebook
import pandas as pd

def domains():
    """domains prints the current potential local domains as a table to stdout

    [extended_summary]
    """
    cfg = get_config()
    df = pd.DataFrame(cfg['domains'])
    df.dataset = df.dataset.str.title()
    df = df.pivot(columns='domain', values='dataset')
    temp = []
    for col in df:
        temp.append(df[col].drop_duplicates().sort_values().dropna().reset_index(drop=True))
    df = pd.concat(temp, axis=1)
    df = df[df.notnull().sum().sort_values(ascending=False).index]
    df = df.fillna('')
    df.columns = df.columns.str.upper()
    if isnotebook:
        return df
    else: 
        print(df.to_string())    
    



from pathlib import Path
import json
from functools import wraps
import time
import datetime as dt
from scipy.stats import circmean,circstd
from scipy.signal import savgol_filter
import pyarrow.parquet as pq
import dateutil
import os
import pyarrow as pa
import requests
import yaml
from IPython import get_ipython
import zipfile
import pandas as pd
import numpy as np


def segment_freq(v:pd.Series,
                 window_length:int=59,
                 polyorder:int=1,
                 r:int=10,
                 lb:float=0.0,
                 ub:float=360.0,
                 step:float=0.25)-> pd.DataFrame:
    """segment_freq [summary]

    [extended_summary]

    Args:
        v (pd.Series): [description]
        window_length (int, optional): [description]. Defaults to 59.
        polyorder (int, optional): [description]. Defaults to 1.
        r (int, optional): [description]. Defaults to 10.

    Returns:
        pd.DataFrame: [description]
    """
    ix = np.arange(lb,ub,step)
    vc = v.value_counts(normalize=True).sort_index().rename('freq').to_frame()
    vc = vc.reindex(ix)
    vc = vc.assign(freq_smooth = savgol_filter(vc.freq, window_length, polyorder))
    x = vc.freq_smooth.values
    vc = vc.assign(freq_cumsum=vc.freq_smooth.cumsum())
    L,idx,m = x**0-1,np.argsort(-x),1
    d,n = np.mgrid[-r:r],idx.shape[0]-1
    for i in idx:
        ix = np.clip(i+d,0,n)
        local = L[ix][L[ix]>0]
        labels, counts = np.unique(local, return_counts=True)
        if local.size==0: L[i],m=m,m+1
        if labels.size==1:L[i]=labels[0]
        if labels.size>1:L[i]=labels[np.argmax(counts)]
    vc = vc.assign(labels=L)
    return vc


def segment_summary(vc,shift=180):
    """segment_summary [summary]

    [extended_summary]

    Args:
        vc ([type]): [description]
    """
    def f(x): return dt.time(*angles_to_time(x))
    summary = (vc.reset_index().
                    groupby('labels').
                    agg(start=('index','first'),
                        end=('index','last'),
                        proportion=('freq','sum')))
    summary.start = ((summary.start-shift)%360).transform(f)
    summary.end = ((summary.end-shift)%360).transform(f)
    summary.proportion = (summary.proportion*100).round(1)
    return summary

def timer(desc : str = None):
    """timer is a wrapper decorator to report functions duration
    Args:
        desc (str, optional): [description line to print to sdout]. Defaults to None.
    """
    def wrapper(fun):
        @wraps(fun)
        def wrapped(*fun_args, **fun_kwargs):
            start = time.perf_counter()
            if len(fun_args)>0 and isinstance(fun_args[0], str):
                prefix = f'Finished {desc} {fun_args[0]} in:'
            else:
                prefix = f'Finished {desc} in:'
            out = fun(*fun_args, **fun_kwargs)
            elapsed = time.perf_counter() - start
            dur = f'{np.round(elapsed,1)}'
            print(f"{prefix:<40}{dur:>10} {'seconds':<10}")
            return out
        return wrapped
    return wrapper


class BearerAuth(requests.auth.AuthBase):
    """BearerAuth manages the coupling of a token to requests framework

    Args:
        requests ([type]): [description]
    """
    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers["authorization"] = "Bearer " + self.token
        return r


def isnotebook() -> bool:
    """isnotebook checks if the run environment is a jupyter notebook


    Returns:
        bool: [description]
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True
        elif shell == 'TerminalInteractiveShell':
            return False
        else:
            return False
    except NameError:
        return False


def date2iso(date: str, output_fmt: str ='%Y-%m-%dT%H:%M:%S.%f'):
    """date2iso convert a date string to iso format 

    Args:
        date (str): [description]
        output_fmt (str, optional): [description]. Defaults to '%Y-%m-%dT%H:%M:%S.%f'.

    Returns:
        [type]: [description]
    """
    dt = dateutil.parser.parse(date)
    return dt.strftime(output_fmt)+'Z'


def read_table(filename: str, columns: list = None) -> pd.DataFrame:
    """read_table reads a parquet pyspark file 

    Args:
        filename (str): filename either in relative or in absoulte path 
        columns (list, optional): specific columns to load. Defaults to None.

    Returns:
        pd.DataFrame: [description]
    """
    if columns is None:
        data = pq.read_table(filename).to_pandas()
    else:
        data = pq.read_table(filename, columns=columns).to_pandas()
    return data


def update_table(data: pd.DataFrame,
                 filename: str,
                 compression: str,
                 meta_content: dict = {}) -> None:
    """update_table updates a parquet pyspark file 

    Args:
        data (pd.DataFrame): [description]
        filename (str): [description]
        compression (str): [description]
        meta_content (dict, optional): [description]. Defaults to {}.
    """
    table = pa.Table.from_pandas(data)
    table = inject_metadata(table, meta_content)
    pqwriter = pq.ParquetWriter(
        filename, table.schema, compression=compression)
    pqwriter.write_table(table=table)
    pqwriter.close()


def inject_metadata(table: pa.table, meta_content: dict) -> pa.Table:
    """inject_metadata replaces metadata in a parquet pyspark file 

    [extended_summary]

    Args:
        table (pa.table): [description]
        meta_content (dict): [description]

    Returns:
        pa.Table: [description]
    """
    pandas_meta = table.schema.metadata
    metadata = json.dumps(meta_content)
    metadata = {
        'minder'.encode(): metadata.encode(),
        **pandas_meta
    }
    table = table.replace_schema_metadata(metadata)
    return table


def write_table(data: pd.DataFrame,
                filename: str,
                compression: str,
                meta_content: dict = {}) -> None:
    """write_table writes data into a parquet pyspark file 

    [extended_summary]

    Args:
        data (pd.DataFrame): [description]
        filename (str): [description]
        compression (str): [description]
        meta_content (dict, optional): [description]. Defaults to {}.
    """
    if filename.split('.')[-1] != 'parquet':
        filename = f'{filename}.parquet'
    table = pa.Table.from_pandas(data)
    table = inject_metadata(table, meta_content)
    pq.write_table(table, filename, compression=compression)


def read_metadata(filename: str):
    """read_metadata return only the metadata from a parquet pyspark file 

    Args:
        filename (str): [description]

    Returns:
        [type]: [description]
    """
    return pq.read_metadata(filename)


def write_yaml(local_file:str, data:dict):
    """write_yaml writes a dictionary structure into a yaml file

    Args:
        local_file (str): [description]
        data (dict): [description]
    """

    with open(local_file, 'w') as yamlfile:
        yaml.safe_dump(data, yamlfile)


def update_yaml(local_file: str, data: dict):
    """update_yaml updates a dictionary structure onto a yaml file

    [extended_summary]

    Args:
        local_file (str): [description]
        data (dict): [description]
    """
    with open(local_file, 'r') as yamlfile:
        currdata = yaml.safe_load(yamlfile)
        data_merged = dict(merge_dicts(currdata, data))
    write_yaml(local_file, data_merged)


def load_csv_from_zip(_zip, csv_file):
        df = pd.read_csv(_zip.open(csv_file),
                         encoding='unicode_escape', low_memory=False)
        return df

def load_yaml(local_file: str) -> dict :
    """load_yaml loads a yaml file into a dictionary

    Args:
        local_file (str): [description]

    Returns:
        dict: [description]
    """
    with open(local_file, 'r') as yamlfile:        
        return yaml.safe_load(yamlfile)


def path_exists(local_file:str):
    """path_exists checks if a file exists in the local filesystem

    [extended_summary]

    Args:
        local_file (str): [description]

    Returns:
        [type]: [description]
    """
    return Path(local_file).exists()


def set_path(local_file: str):
    """set_path checks if a parent folder exists and if not creates it 
    Args:
        local_file (str): [description]
    """
    if not path_exists(local_file):
        Path(Path(local_file).absolute()
             .parent).mkdir(parents=True, exist_ok=True)

def merge_dicts(d1:dict, d2:dict):
    """merge_dicts merges two dictionaries

    Args:
        d1 (dict): [description]
        d2 (dict): [description]

    Yields:
        [type]: [description]
    """
 
    for k in set(d1) | set(d2):
        if k in d1 and k in d2:
            if isinstance(d1[k], dict) and isinstance(d2[k], dict):
                yield (k, dict(merge_dicts(d1[k], d2[k])))
            else:
                yield (k, d2[k])
        elif k in d1:
            yield (k, d1[k])
        else:
            yield (k, d2[k])


def lagged_df(df:pd.DataFrame, factor:str='activity', lags:int=7):
    """lagged_df returns a lagged dataframe for a specific factor 

    [extended_summary]

    Args:
        df (pd.DataFrame): [description]
        factor (str, optional): [description]. Defaults to 'activity'.
        lags (int, optional): [description]. Defaults to 7.

    Returns:
        [type]: [description]
    """
    return pd.concat([df[factor].shift(lag).rename(f't(-{lags-lag-1})') for lag in range(lags)], axis=1).dropna()


def reindex_ts(x:pd.Series, freq:str):
    """reindex_ts reindex a timeseries by some freq 

    Args:
        x (pd.Series): [description]
        freq (str): goes in the time format of pandas 

    Returns:
        [type]: [description]
    """
    rng = pd.date_range(x.index.min(), x.index.max(), freq=freq)
    _df = pd.DataFrame(x, index=x.index).reindex(rng)
    return _df


def load_zip_csv(zip_file:str, csv_file:str) -> pd.DataFrame:
    """load_zip_csv returns a specific csv file from a zip file 

    Args:
        zip_file (str): [description]
        csv_file (str): [description]

    Returns:
        pd.DataFrame: [description]
    """
    zip = zipfile.ZipFile(zip_file)
    df = pd.read_csv(zip.open(csv_file),
                     encoding='unicode_escape',
                     low_memory=False)
    return df


def shift_row_to_bottom(df:pd.DataFrame, index_to_shift:str)-> pd.DataFrame:
    """shift_row_to_bottom shifts a specific index to the bottom of a dataframe
    Args:
        df (pd.DataFrame): [description]
        index_to_shift (str): [description]

    Returns:
        pd.DataFrame: [description]
    """
    idx = [i for i in df.index if i != index_to_shift]
    return df.loc[idx+[index_to_shift]]


def shift_row_to_top(df: pd.DataFrame, index_to_shift: str) -> pd.DataFrame:
    """shift_row_to_top shifts a specific index to the top of a dataframe

    Args:
        df (pd.DataFrame): [description]
        index_to_shift (str): [description]

    Returns:
        pd.DataFrame: [description]
    """
    idx = [i for i in df.index if i != index_to_shift]
    return df.loc[[index_to_shift]+idx]


def localize_time(df:pd.DataFrame, factors:list, timezones=None):
    """localize_time [summary]

    [extended_summary]

    Args:
        df (pd.DataFrame): [description]
        factors (list): [description]
        timezones ([type], optional): [description]. Defaults to None.

    Returns:
        [type]: [description]
    """
    data = []
    if timezones is None:
        timezones = ['Europe/London']
        df['timezone'] = np.repeat('Europe/London', df.shape[0])
    for tz in timezones:
        _df = df[df.timezone == tz].copy()
        try:
            for factor in factors:
                dt = pd.to_datetime(_df[factor]).dt.tz_localize(None)
                offset = pd.Series([t.utcoffset() for t in dt.dt.tz_localize(
                    tz, ambiguous=True, nonexistent='shift_forward')], index=dt.index)
                _df[factor] = dt + offset
            data.append(_df)
        except:
            print(tz)
    data = pd.concat(data)
    return data

def between_time(df,factor,start_time,end_time):
    index = pd.DatetimeIndex(df[factor])
    return df.iloc[index.indexer_between_time(start_time,end_time)]


def epoch_to_local(dt: pd.Series, tz: str = 'Europe/London', unit: str = 's', shift: int = 0)-> pd.Series:
    """epoch_to_local converts epoch

    [extended_summary]

    Args:
        dt (pd.Series): [description]
        tz (str, optional): [description]. Defaults to 'Europe/London'.
        unit (str, optional): [description]. Defaults to 's'.
        shift (int, optional): [description]. Defaults to 0.

    Returns:
        pd.Series: [description]
    """
    dt = pd.to_datetime(dt, unit=unit)
    offset = pd.Series([t.utcoffset() for t in dt.dt.tz_localize(
        tz, ambiguous=True, nonexistent='shift_forward')], index=dt.index)
    return dt + offset + pd.Timedelta(hours=shift)

def utc_to_local(dt:pd.Series, tz:str='Europe/London', shift:int=-2):
    """utc_to_local converts a timeseries from utc to a specific timezone 

    Args:
        dt (pd.Series): [description]
        tz (str, optional): [description]. Defaults to 'Europe/London'.
        shift (int, optional): [description]. Defaults to -2.

    Returns:
        [type]: [description]
    """
    dt = pd.to_datetime(dt).dt.tz_localize(None)
    offset = pd.Series([t.utcoffset() for t in dt.dt.tz_localize(
        tz, ambiguous=True, nonexistent='shift_forward')], index=dt.index)
    return dt + offset + pd.Timedelta(hours=shift)


def time_cdf(times:pd.Series, name:str)->pd.Series:
    """time_cdf return a cdf of times as a pandas Series

    Args:
        times (pd.Series): [description]
        name (str): [description]

    Returns:
        pd.Series: [description]
    """
    vc = times.value_counts(normalize=True).sort_index()
    rng = pd.date_range('2021-01-01 12:00:00',
                        '2021-01-02 12:00:00', freq='1T')
    vc = vc.reindex(rng.time)
    vc = vc.fillna(0).cumsum()
    vc.name = name
    return vc


def process_transition(df:pd.DataFrame, groupby:list, datetime:str, value:str, covariates=None) -> pd.DataFrame:
    """process_transition convert a timeseries DataFrame with datetimes to a transition dataframe 

    Args:
        df (pd.DataFrame): [description]
        groupby (list): [description]
        datetime (str): [description]
        value (str): [description]
        covariates ([type], optional): [description]. Defaults to None.

    Returns:
        [type]: [description]
    """
    data = []
    df = df.sort_values(datetime)
    for ix, subset in df.groupby(groupby):
        if len(groupby) > 1:
            index = pd.MultiIndex.from_tuples(
                [ix for _ in range(subset.shape[0])], names=groupby)
        else:
            index = np.tile(ix, subset.shape[0])
        subset = subset.reset_index(drop=True)
        dur = (subset[datetime].shift(-1) - subset[datetime]
               ).dt.total_seconds().rename('dur')
        start_date = subset[datetime].rename('start_date')
        end_date = subset[datetime].shift(-1).rename('end_date')

        transition = (subset[value].astype(
            str) + '>' + subset[value].shift(-1).astype(str)).rename('transition')
        if covariates is None:
            subset = pd.concat([start_date, end_date, transition, dur], axis=1)
        else:
            cov = subset[covariates]
            subset = pd.concat(
                [start_date, end_date, cov, transition, dur], axis=1)
        subset.index = index
        subset.index.names = groupby
        data.append(subset.dropna())
    data = pd.concat(data)
    dtypes = {'start_date': 'datetime64',
              'end_date': 'datetime64',
              'transition': 'category',
              'dur': 'float'}
    return data.astype(dtypes)


def time_to_angles(time, day=24*60**2):
    """time_to_angles [summary]

    [extended_summary]

    Args:
        time ([type]): [description]
        day ([type], optional): [description]. Defaults to 24*60**2.

    Returns:
        [type]: [description]
    """
    time =  str_to_time(time) if str is type(time) else time
    return 360*(time.hour*60**2+time.minute*60+time.second)/day 


def str_to_time(time):
    """str_to_time [summary]

    [extended_summary]

    Args:
        time ([type]): [description]

    Returns:
        [type]: [description]
    """
    return dt.time(*[int(t) for t in time.split(':')])

def seconds_to_time(seconds):
    """seconds_to_time [summary]

    [extended_summary]

    Args:
        seconds ([type]): [description]

    Returns:
        [type]: [description]
    """
    if pd.isnull(seconds):
        return np.full(3,np.nan)
    else:
        h, m = np.divmod(seconds, 3600)
        m, s = np.divmod(m, 60)
        hms = np.array([h, m, s]).astype(int)
        return hms.T

def angles_to_time(angles, day=24*60**2):
    """angles_to_time [summary]

    [extended_summary]

    Args:
        angles ([type]): [description]
        day ([type], optional): [description]. Defaults to 24*60**2.

    Returns:
        [type]: [description]
    """
    return seconds_to_time((angles * day) / 360)

def times_to_angles(times, day=24*60**2):
    """times_to_angles [summary]

    [extended_summary]

    Args:
        times ([type]): [description]
        day ([type], optional): [description]. Defaults to 24*60**2.

    Returns:
        [type]: [description]
    """
    return np.array([time_to_angles(t,day) for t in times])    

def std_time(times,kind='time'):
    """std_time [summary]

    [extended_summary]

    Args:
        times ([type]): [description]
        kind (str, optional): [description]. Defaults to 'time'.

    Returns:
        [type]: [description]
    """
    day = 24*60**2
    angles = times_to_angles(times, day)
    sd_angle = circstd(angles,high=360)
    if kind == 'time':
        return dt.time(*angles_to_time(sd_angle))#np.timedelta64(sd_seconds,'s')
    if kind == 'angles':
        return sd_angle
    if kind == 'datetime':
        return dt.datetime(2021,1,1,*angles_to_time(sd_angle))
    if kind == 'timedelta':
        return dt.timedelta(seconds=(sd_angle * day) / 360)
   
def mean_time(times,kind='time'):
    """mean_time [summary]

    [extended_summary]

    Args:
        times ([type]): [description]
        kind (str, optional): [description]. Defaults to 'time'.

    Returns:
        [type]: [description]
    """
    day = 24*60**2
    angles = times_to_angles(times, day)
    mean_angle = circmean(angles,high=360)
    if kind == 'time':
        return dt.time(*angles_to_time(mean_angle))#np.timedelta64(sd_seconds,'s')
    if kind == 'angles':
        return mean_angle
    if kind == 'datetime':
        return dt.datetime(2021,1,1,*angles_to_time(mean_angle))
    if kind == 'timedelta':
        return dt.timedelta(seconds=(mean_angle * day) / 360)



def shift_row_to_bottom(df, index_to_shift):
    """shift_row_to_bottom [summary]

    [extended_summary]

    Args:
        df ([type]): [description]
        index_to_shift ([type]): [description]

    Returns:
        [type]: [description]
    """
    idx = [i for i in df.index if i!=index_to_shift]
    return df.loc[idx+[index_to_shift]]

def shift_row_to_top(df, index_to_shift):
    """shift_row_to_top [summary]

    [extended_summary]

    Args:
        df ([type]): [description]
        index_to_shift ([type]): [description]

    Returns:
        [type]: [description]
    """
    idx = [i for i in df.index if i!=index_to_shift]
    return df.loc[[index_to_shift]+idx]


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dcarte']

package_data = \
{'': ['*'], 'dcarte': ['source_yaml/*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'ipykernel>=6.6.0,<7.0.0',
 'ipython>=7.30.1,<8.0.0',
 'ipywidgets>=7.6.5,<8.0.0',
 'jupyter>=1.0.0,<2.0.0',
 'lab>=7.0,<8.0',
 'numpy>=1.21.5,<2.0.0',
 'pandas>=1.3.5,<2.0.0',
 'pyarrow>=3.0.0,<6.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'requests>=2.26.0,<3.0.0',
 'scipy>=1.7.3,<2.0.0',
 'tqdm>=4.62.3,<5.0.0']

setup_kwargs = {
    'name': 'dcarte',
    'version': '0.3.1',
    'description': 'DCARTE is a dataset ingestion tool from DCARTE UK-DRI CAre Research and TEchnology',
    'long_description': None,
    'author': 'eyal soreq',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



from dataclasses import dataclass, field
import datetime as dt
import json
import os
from pathlib import Path
import pandas as pd
from time import sleep
from io import StringIO
import requests
from .config import get_config
from .utils import (write_table,
                   read_table,
                   read_metadata,
                   isnotebook,
                   date2iso,
                   BearerAuth,
                   path_exists,
                   timer,
                   set_path)

if isnotebook():
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

cfg = get_config()
NOW = date2iso(str(dt.datetime.now()))

@dataclass
class MinderDataset(object):
    """MinderDataset class handles the downloading of datasets from the minder reserch platform

    [extended_summary]

    Args:
        dataset_name ([str]): [description]
        datasets ([list]): [description]
        columns ([list]): [description]
        domain ([str]): [description]
        dtypes ([list]): [description]
        since ([list]): [description]
        until ([list]): [description]
        delay ([list]): [description]
        auth ([list]): [description]
        headers ([list]): [description]
        server ([list]): [description]
        token ([list]): [description]
        compression ([list]): [description]
        data_folder ([list]): [description]
        data ([list]): [description]
        request_id ([list]): [description]
        reload ([list]): [description]
        reapply ([list]): [description]
        update ([list]): [description]

    Raises:
        Exception: [description]

    Returns:
        [type]: [description]
    """
    dataset_name: str
    datasets: list
    columns: list
    domain: str
    dtypes: list
    since: str = '2019-04-01'
    until: str = NOW
    delay: float = 1
    headers: dict = field(default_factory=lambda: cfg['headers'])
    server: str = cfg['server']
    home: Path = cfg['home']
    compression: str = cfg['compression']
    data_folder: str = cfg['data_folder']
    data: pd.DataFrame = field(default_factory=lambda: pd.DataFrame())
    request_id: str = ''
    reload: bool = False
    reapply: bool = False
    update: bool = False

    def __post_init__(self):
        self._delay = dt.timedelta(hours=self.delay)
        self.since = date2iso(self.since)
        self.until = date2iso(self.until)
        self.auth = BearerAuth(os.getenv('MINDER_TOKEN'))
        self.data_request = {'since': self.since,
                             'until': self.until,
                             'datasets': {ds: {"columns": self.columns}
                                          for ds in self.datasets}}
        self.local_file = (f'{self.data_folder}/'
                           f'{self.domain}/'
                           f'{self.dataset_name}.parquet')

        if not path_exists(self.local_file) or self.reload:
            set_path(self.local_file)
            self.download_dataset()
        elif self.update:
            self.update_dataset()
        else:
            self.data = read_table(self.local_file)
    
    def download_dataset(self):
        self.post_request()
        self.process_request()
        self.download_data()
        self.update_metadata()
        self.save_dataset()

    def post_request(self):
        request = requests.post(self.server,
                                data=json.dumps(self.data_request),
                                headers=self.headers,
                                auth=self.auth)
        if request.status_code != 403:
            self.request_id = (request.headers['Content-Location']
                               .split('/')[-1])
        else:
            raise Exception("You need an active VPN connection")

    def process_request(self, sleep_time:int=10):
        request_output = self.get_output()
        print(f'Processing {self.dataset_name} ',end=':')
        while request_output.empty:
            sleep(sleep_time)
            request_output = self.get_output()
        self.csv_url = request_output

    def get_output(self):
        request = requests.get(self.server, auth=self.auth)
        request_elements = pd.DataFrame(request.json())
        request_elements = request_elements[request_elements.id ==
                                            self.request_id]
        output = pd.DataFrame()
        if not request_elements.empty:
            if request_elements.jobRecord.notnull().iat[0]:
                output = pd.DataFrame(
                    request_elements.jobRecord.values[0]['output'])
            if request_elements.status.iat[0] == 202:
                print('*',end='')
        return output
    
    def download_data(self):
        data = []
        for idx, url in enumerate(tqdm(self.csv_url.url,
                                       desc=f'Downloading {self.dataset_name}',
                                       dynamic_ncols=True)):
            request = requests.get(url, stream=True, auth=self.auth)
            decoded_data = StringIO(request.content.decode('utf-8-sig'))
            df = pd.read_csv(decoded_data, sep=',', engine='python')
            df['source'] = self.csv_url.type[idx]
            data.append(df)
        self.data = pd.concat(data).reset_index(drop=True)
        if (self.data[self.columns[0]] == self.columns[0]).any():
            self.data = self.data[self.data[self.columns[0]]
                                  != self.columns[0]]
        dtypes = dict(zip(self.columns, self.dtypes))
        self.data = self.data.replace({'false':0.0,'true':1.0}).astype(dtypes)

    def update_metadata(self):
        if path_exists(self.local_file):
            since = self.load_metadata()['since']
        else:
            since = self.since
        self.metadata = {'since': since,
                         'until': self.until,
                         "request_id": self.request_id,
                         'Mac': cfg['mac']}

    def load_metadata(self):
        hdr = read_metadata(self.local_file)
        metadata = json.loads(hdr.metadata[b'minder'].decode())
        return metadata

    def append_dataset(self):
        # TODO: fix the append to file function
        dtypes = dict(zip(self.columns, self.dtypes))
        _data = read_table(self.local_file).replace({'false':0.0,'true':1.0}).astype(dtypes)
        write_table(pd.concat([_data, self.data]).reset_index(drop=True),
                    self.local_file,
                    self.compression,
                    self.metadata)

    def save_dataset(self):
        dtypes = dict(zip(self.columns, self.dtypes))
        write_table(self.data.astype(dtypes),
                    self.local_file,
                    self.compression,
                    self.metadata)

    def update_dataset(self):
        hdr = read_metadata(self.local_file)
        metadata = json.loads(hdr.metadata[b'minder'].decode())
        until = pd.to_datetime(metadata['until']).tz_localize(None) + self._delay
        if until < pd.to_datetime(dt.datetime.now()):
            self.data_request['since'] = metadata['until']
            self.data_request['until'] = self.until
            self.post_request()
            self.process_request()
            self.download_data()
            self.update_metadata()
            self.append_dataset()
            self.load_dataset()
        else:
            self.load_dataset()

    def load_dataset(self):
        self.data = read_table(self.local_file)
